



pybamm_edit_parameter --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
pybamm_add_parameter --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
pybamm_rm_parameter --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
pybamm_install_jax --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
